import os
from pathlib import Path

import pytest

from godkjenn.artifact import Artifact
import godkjenn.config
from godkjenn.verification import MismatchError, verify
from godkjenn.environment import Environment


def pytest_addoption(parser):
    parser.addini('godkjenn_config', 'Path to godkjenn configuration file')


INSTRUCTIONS = """{message}

If you wish to accept the received result, run:

    godkjenn {options} accept "{test_id}"
"""


def _make_instructions(message, test_id, pytest_config):
    """Calculate 'accept' instructions.

    Args:
        message: Any message to put before the instructions.
        test_id: The string id of the test.
        pytest_config: The full pytest configuration object.

    Returns: A string representing a command to run that will accept the received data.
    """
    pt_root = pytest_config.rootdir.relto(os.getcwd())
    gk_path = Path(pt_root) / pytest_config.getini('godkjenn_config')

    if gk_path.exists() and gk_path.is_file():
        options = "--config={}".format(gk_path)
    else:
        options = "--root-dir={}".format(pt_root)

    return INSTRUCTIONS.format(
        message=message,
        test_id=test_id,
        options=options)


def pytest_runtest_makereport(item, call):
    """pytest hook that runs to generate specialized reports.

    We find all Approver fixture arguments and generate a test report containing instructions on how to proceed (e.g.
    update the accepted value if desired).
    """
    import _pytest.runner

    if call.when == 'call' and call.excinfo is not None:
        exc = call.excinfo.value

        # The test may have failed for reasons besides verification failure.
        if not isinstance(exc, MismatchError):
            return

        # There are potentially more than one approver, so we loop over them.
        for approver in _approver_args(item):
            instructions = _make_instructions(
                exc.message,
                approver._test_id,
                item.config)
            return _pytest.runner.TestReport(
                location=item.location,
                keywords=item.keywords,
                outcome='failed',
                when=call.when,
                nodeid=item.nodeid,
                longrepr=instructions,
            )


@pytest.fixture(scope='session')
def godkjenn_env(pytestconfig):
    """Get the godkjenn configuration.

    If the pytest.ini specifies a config file, we'll try to use that. If it doesn't, we'll look for 'godkjenn.toml' in
    the root directory. If either exists, we load it (as TOML) and return the 'godkjenn' section. If there is no
    'godkjenn' section, we'll return a default configuration.
    """
    config_path = pytestconfig.getini('godkjenn_config')
    env = Environment()

    if not config_path:
        config = env.default_config(Path(pytestconfig.rootdir))
    else:
        config_path = Path(str(pytestconfig.rootdir / config_path))
        config = godkjenn.config.load_config(config_path)

    env.activate(config)
    return env


@pytest.fixture(name='godkjenn')
def godkjenn_fixture(request, godkjenn_env):
    "Returns an object on which you can call `verify()`."
    test_id = request.node.nodeid
    return Approver(test_id, godkjenn_env)


class Approver:
    """Type returned from the godkjenn fixture.

    Users can call the `verify()` method to check their latest results.
    """

    def __init__(self, test_id, env):
        self._test_id = test_id
        self._env = env

    def verify(self, received, artifact_config={}):
        """Check the latest received data against the accepted version.

        If there's a mismatch, this will trigger a pytest failure (i.e. via `assert`).

        Args:
            received: The received test data to be compared with the approved.
        """
        verify(self._env, self._test_id,
               Artifact(received, artifact_config))


def _approver_args(item):
    "Find all Approver fixture arguments to a test item."
    for arg in item.funcargs.values():
        if isinstance(arg, Approver):
            yield arg



"""Implementation of the 'environment' concept.
"""

import os
from pathlib import Path

import fitb
from godkjenn.config import Config


class Environment:
    """Manages extensions and the overall program configuration.
    """
    def __init__(self):
        self._extension_points = fitb.load_from_pkg_resources('godkjenn')
        self._config = None

    @property
    def config(self):
        if not self.activated:
            raise ValueError('Environment is not yet activated.')
        return self._config

    @property
    def activated(self):
        return self._config is not None

    def activate(self, config):
        self._config = config
        for point in self._extension_points.values():
            point.activate(config)

    def default_config(self, root_path=None):
        if root_path is None:
            root_path = Path(os.getcwd())

        config = Config(
            root_path,
            {
                'vault_type': 'fs-vault',
                'differ_type': 'text',
                'comparator_type': 'exact',
            }
        )
        extension_config = fitb.default_config(*self._extension_points.values())
        fitb.merge(config, extension_config)

        return config

    def get_vault(self):
        """Construct a vault.

        This reads a config dict to determine how to create a vault.

        This will look for the "vault_type" key in the config, defaulting to 'fs-vault'. This indicates the name of the
        vault plugin to use.

        It will then look for the subdict with the same name as the vault type, defaulting to an empty dict if this is
        not found. This subconfig dict is passed to the plugin driver.

        Args: config: A config dict.

        Returns: A vault instance.
        """
        vault_plugin_name = self.config['vault_type']
        extension_point = self._extension_points['vault']
        return extension_point[vault_plugin_name]

    def get_comparator(self):
        """Construct a comparator object.

        Args:
            name: Name of comparator type to get.
            config: A config dict.

        Returns: A comparator instance.
        """
        plugin_name = self.config['comparator_type']
        extension_point = self._extension_points['comparator']
        return extension_point[plugin_name]

    def comparator_names(self):
        """Get all comparator plugin names.

        Returns: An iterable of comparator names.
        """
        extension_point = self._extension_points['comparator']
        return extension_point.names()

    def get_differ(self):
        """Construct a comparator object.

        Args:
            name: Name of comparator type to get.
            config: A config dict.

        Returns: A comparator instance.
        """
        plugin_name = self.config['differ_type']
        extension_point = self._extension_points['differ']
        return extension_point[plugin_name]

    def differ_names(self):
        """Get all differ plugin names.

        Returns: An iterable of differ plugin names.
        """
        extension_point = self._extension_points['differ']
        return extension_point.names()



"""Simple file-system based implementation of a vault.
"""

from enum import Enum
from pathlib import Path

import fitb
import godkjenn.config
from godkjenn.artifact import Artifact


class FSVault:
    """File-system implementation of a vault.

    This just keeps accepted data in files with the "accepted" suffix and received data in files suffixed with
    "received". Id's in this vault are simply paths.

    This assumes it has complete control over the files under `root_directory`.

    Args:
        root_path: The root directory under which this will store received and accepted data.
    """

    def __init__(self, root_path: Path):
        self._root_path = root_path

    @property
    def root_path(self):
        "Root path of the vault."
        return self._root_path

    def accepted(self, test_id):
        """Get the current accepted value for `test_id`.

        Args:
            test_id: ID of the test.

        Returns: An Artifact instance.

        Raises:
            KeyError: `test_id` does not have accepted data in the vault.
        """
        return self._get(test_id, _Kind.accepted)

    def accept(self, test_id):
        """Accept the current received data for `test_id`.

        Args:
            test_id: The ID of the test to accept.

        Raises:
            KeyError: There is no received data for `test_id`.
        """
        artifact = self.received(test_id)

        self._put(test_id, _Kind.accepted, artifact)

        p = self._full_path(test_id, _Kind.received)
        if p.exists():
            p.unlink()

    def received(self, test_id):
        """Get the current received value for `test_id`.

        Args:
            test_id: ID of test.

        Returns: An Artifact describing the received data `test_id`.

        Raises:
            KeyError: There is no received data for `test_id`.
        """
        return self._get(test_id, _Kind.received)

    def receive(self, test_id, artifact):
        """Set new received data for a test.

        Args:
            test_id: ID of the test for which to receive data.
            artifact: Artifact describing received data for the test.
        """
        self._put(test_id, _Kind.received, artifact)

    def ids(self):
        """Get all IDs in the vault.

        This is all IDs that have either or both of accepted and received data. There is no
        order to the results. Each ID is included only once in the output, even if it has both
        received and accepted data.

        Returns: An iterable of all test IDs.
        """
        return set(
            str(path.parent.relative_to(self._root_path))
            for kind in _Kind
            for path in self._root_path.glob('**/{}'.format(kind.value)))

    def _full_path(self, path, kind):
        return self._root_path / path / kind.value

    def _get(self, path, kind):
        full_path = self._full_path(path, kind)

        if not full_path.exists():
            raise KeyError('no {} data: {}'.format(kind.value, path))

        with full_path.open(mode='rb') as handle:
            data = handle.read()

        config = godkjenn.config.load_config(
            full_path.parent / 'config')

        return Artifact(data, config)

    def _put(self, path, kind, artifact):
        full_path = self._full_path(path, kind)

        full_path.parent.mkdir(parents=True, exist_ok=True)

        with full_path.open(mode='wb') as handle:
            handle.write(artifact.data)

        godkjenn.config.save_config(
            artifact.config,
            full_path.parent / 'config')


class _Kind(Enum):
    accepted = 'accepted'
    received = 'received'


def extension(extension_point):
    extension_point.add(
        name='fs-vault',
        description='Simple filesystem-based vault',
        config_options=(fitb.option.Option('dir_name', 'Name of vault directory', '.godkjenn'),),
        activate=lambda full, ext: FSVault(root_path=full.root_path / ext['dir_name']))



class Artifact:
    """A single received or accepted piece of data.

    An artifact includes the byte data itself as well as
    any extra configuration it uses.
    """
    def __init__(self, data, config):
        self._config = config
        self._data = data

    @property
    def config(self):
        "Extra configuration for this data."
        return self._config

    @property
    def data(self):
        "Data for the object (bytes)."
        return self._data

    def __eq__(self, rhs):
        return self.data == rhs.data and self.config == rhs.config



"""Support for loading/getting configuration dictionaries.
"""

import toml


class Config(dict):
    """Dict subclass with a root-path attribute.
    """

    def __init__(self, root_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._root_path = root_path

    @property
    def root_path(self):
        return self._root_path


def deserialize_config(config_string, root_path):
    config = toml.loads(config_string)
    return Config(root_path, config.get('godkjenn', {}))


def load_config(path):
    return deserialize_config(
        path.read_text(encoding='utf-8'),
        path.parent)


def serialize_config(config):
    return toml.dumps({'godkjenn': dict(config)})


def save_config(config, path):
    config_string = serialize_config(config)
    path.write_text(config_string, encoding='utf-8')



def compare(accepted, received):
    "Exact byte-wise comparison."
    return accepted == received


def extension(extension_point):
    extension_point.add(
        name='exact',
        description='Exact byte-wise comparison',
        activate=lambda _full, _ext: compare
    )



from pathlib import Path

from setuptools import find_packages, setup

setup(
    name='godkjenn',
    version='0.5.0',
    packages=find_packages('src'),

    author='Sixty North AS',
    author_email='austin@sixty-north.com',
    description='Approval testing for Python3',
    license='MIT',
    keywords='',
    url='http://bitbucket.org/sixty-north/godkjenn',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Framework :: Pytest',
    ],
    platforms='any',
    include_package_data=True,
    package_dir={'': 'src'},
    install_requires=[
        'click',
        'exit-codes',
        'toml',
    ],
    extras_require={
        'dev': ['bumpversion', 'twine'],
        'test': ['hypothesis', 'pytest'],
        'pytest-plugin': ['pytest'],
    },
    entry_points={
        'pytest11': [
            'godkjenn = godkjenn.integration.pytest_plugin',
        ],
        "godkjenn.vault": [
            'fs-vault = godkjenn.vaults.fs_vault:extension',
        ],
        'godkjenn.comparator': [
            'exact = godkjenn.comparators.exact:extension',
        ],
        'godkjenn.differ': [
            'text = godkjenn.differs.text:extension',
            'binary = godkjenn.differs.binary:extension',
        ],
        'console_scripts': [
            'godkjenn = godkjenn.cli:main',
        ]
    },
    long_description=Path('README.rst').read_text(encoding='utf-8'),
)



"""The core approval testing algorithm.

This is where we check the latest *received* data with the latest *accepted* data.
"""


def verify(env, test_id, received):
    """Check if `received` matches the current accepted value for the test_id.

    If `received` doesn't match the accepted value, this will raise MismatchError§.

    Args:
        env: An activated Environment.
        test_id: The ID of the test that produced `received`.
        received: An Artifact representing the received data.
    """
    vault = env.get_vault()

    try:
        accepted = vault.accepted(test_id)

        # TODO: Create "complete" config by merging accepted.config into config. See fitb. Use this full config for
        # creating comparator.
        comparator = env.get_comparator()

        if comparator(accepted.data, received.data):
            return

        message = "Received data does not match accepted"
    except KeyError:
        accepted = None
        message = "There is no accepted data"

    vault.receive(test_id, received)
    raise MismatchError(message, received, accepted)


class MismatchError(Exception):
    def __init__(self, message, received, accepted):
        super().__init__(message)
        self._received = received
        self._accepted = accepted

    @property
    def message(self):
        return self.args[0]

    @property
    def received(self):
        return self._received

    @property
    def accepted(self):
        return self._accepted



import difflib


def binary_diff(accepted, received):
    diff = difflib.diff_bytes(
        difflib.unified_diff,
        [accepted.data],
        [received.data])
    return b''.join(diff)


def extension(extension_point):
    extension_point.add(
        name='binary',
        description='Differ for binary files.',
        activate=lambda _full, _ext: binary_diff)



import logging
import sys
from pathlib import Path

import click
from exit_codes import ExitCode

from godkjenn.artifact import Artifact
import godkjenn.config
import godkjenn.environment
from godkjenn.version import __version__

log = logging.getLogger(__name__)


@click.group()
@click.option('--config', help='Config file to load', type=Path)
@click.option('--root-dir', help='Root directory', type=Path)
@click.option('--verbosity',
              default='WARNING',
              help="The logging level to use.",
              type=click.Choice([name
                                 for lvl, name in sorted(logging._levelToName.items())
                                 if lvl > 0],
                                case_sensitive=True))
@click.version_option(version=__version__)
@click.pass_context
def cli(ctx, config, root_dir, verbosity):
    """Command-line interface for godkjenn.
    """
    logging_level = getattr(logging, verbosity)
    logging.basicConfig(level=logging_level)

    env = godkjenn.environment.Environment()

    if config is not None:
        config = godkjenn.config.load_config(config)
    else:
        config = env.default_config(root_dir)

    env.activate(config)

    ctx.obj = env


@cli.command()
@click.argument('test_id')
@click.pass_obj
def accept(env, test_id):
    """Accept the current received data for a test.
    """
    vault = env.get_vault()

    try:
        vault.accept(test_id)
    except KeyError:
        log.error('No received data for {}'.format(test_id))
        return ExitCode.DATA_ERR

    return ExitCode.OK


@cli.command()
@click.pass_obj
def accept_all(env):
    """Accept all received data for a configuration/root directory.
    """
    vault = env.get_vault()

    for test_id in vault.ids():
        try:
            vault.accept(test_id)
        except KeyError:
            # This just means there isn't any received data for the ID, just accepted.
            pass

    return ExitCode.OK


@cli.command()
@click.argument('test_id')
@click.argument('destination', type=click.File(mode='wb'))
@click.pass_obj
def accepted(env, test_id, destination):
    """Get accepted data for a test.
    """
    vault = env.get_vault()

    try:
        artifact = vault.accepted(test_id)
    except KeyError:
        print(f'No accepted data for id {test_id}', file=sys.stderr)
        return ExitCode.DATA_ERR

    destination.write(artifact.data)

    return ExitCode.OK


@cli.command()
@click.argument('test_id')
@click.argument('data_source', type=click.File(mode='rb'))
@click.pass_obj
def receive(env, test_id, data_source):
    """Receive new data for a test.
    """
    vault = env.get_vault()

    data = data_source.read()
    vault.receive(test_id, Artifact(data=data, config={}))

    return ExitCode.OK


@cli.command()
@click.argument('test_id')
@click.argument('destination', type=click.File(mode='wb'))
@click.pass_obj
def received(env, test_id, destination):
    """Get received data for a test.
    """
    vault = env.get_vault()

    try:
        artifact = vault.received(test_id)
    except KeyError:
        print(f'No received data for {test_id}', file=sys.stderr)
        return ExitCode.DATA_ERR

    destination.write(artifact.data)

    return ExitCode.OK


@cli.command()
@click.argument('test_id')
@click.pass_obj
def diff(env, test_id):
    """Get received data for a test.
    """
    vault = env.get_vault()
    try:
        received = vault.received(test_id)
        accepted = vault.accepted(test_id)
    except KeyError:
        print(f'Do not have both received and accepted data. No diff possible.', file=sys.stderr)
        return ExitCode.DATA_ERR

    # TODO: Merge accepted config in
    differ = env.get_differ()

    diff = differ(accepted, received)
    print(diff)

    return ExitCode.OK


@cli.command()
@click.pass_obj
def default_config(env):
    """Print a default config to stdout.
    """
    print(godkjenn.config.serialize_config(
        env.default_config(env.config.root_path)))


@cli.command()
@click.pass_obj
def status(env):
    """Print status of godkjenn vault
    """
    vault = env.get_vault()
    for test_id in vault.ids():
        try:
            accepted = vault.accepted(test_id)
        except KeyError:
            accepted = None

        try:
            received = vault.received(test_id)
        except KeyError:
            received = None

        if accepted is None:
            if received is None:
                assert False, 'Test ID with no information: {}'.format(test_id)
            else:
                message = 'initialized'
        else:
            if received is None:
                message = 'up-to-date'
            elif accepted == received:
                message = 'status-quo'
            else:
                message = 'mismatch'

        print(test_id, message)


def main(argv=None, standalone_mode=True):
    return cli(argv, standalone_mode=standalone_mode)


if __name__ == '__main__':
    sys.exit(main())



import difflib
from functools import partial

import fitb


def text_diff(extension_config, accepted, received):
    # TODO: Merge accepted config
    accepted_text = accepted.data.decode(
        extension_config.get('encoding', 'utf-8'))

    # TODO: Merge received config
    received_text = received.data.decode(
        extension_config.get('encoding', 'utf-8'))

    from io import StringIO
    diff = difflib.unified_diff(
        list(StringIO(accepted_text)),
        list(StringIO(received_text)))

    return ''.join(diff)


def extension(extension_point):
    extension_point.add(
        name='text',
        description='Differ for text files.',
        config_options=(fitb.Option(
            'encoding', 'Encoding of files', 'utf-8'),),
        activate=lambda full, ext: partial(text_diff, ext))
